/// \file mil1553ud_cdev.h
/// \brief инфраструктура для работы с символьными устройствами каналов pci устройства
#ifndef MIL1553UD_CDEV_H
#define MIL1553UD_CDEV_H

#include "mil1553ud_pci.h"

/// \brief имя шаблона класса символьного устройтсва
#define MIL1553_CDEV_CLASS                  "mil1553dev"
/// \brief кол-во символов в имени
#define COUNT_NAME_SYMBOLS                  32
/// \brief невалидное значение MAJOR
#define BAD_DEV_MAJOR                       -1

/// \brief размер блока ДМА (128 байт)
#define SIZE_RAW_DMA_BLOCK                  128
/// \brief мегабайт в байтах (размер кольцевого буфера для одного канала)
#define MEGABYTE                            1024*1024

/// \brief кол-во слов в поддадресе
#define COUNT_WORDS_IN_SA                   32
/// \brief кол-во поддадресов
#define COUNT_SA                            30
/// \brief макрос вычисления поддареса
/// \param num номер по госту
#define SUBA(num)   \
    (num - 1)
/// \brief макрос вычисления номера слова данных
/// \param num номер по госту
#define SUBA_WORD(num) \
    (num - 1)

/// \brief информация о сигналах для прерывания
struct signal_info {
    int target_pid;
    int sig_num;
    struct task_struct* sig_task;
    char isCanSend;
};

/// \brief буфер данных по поддаресам
struct subaddress_data {
    /// \brief данные (массивы поддадресов)
    unsigned short data[COUNT_SA][COUNT_WORDS_IN_SA];
};

/// \brief кол-во блоков в кольцевом буфере
#define INTERRUPT_COUNT_BLOCKS      2048

/// \brief на каждый канал будет создаваться структура кольцевого буфера
struct ring_buffer {
    /// \brief кол-во блоков в кольцевом буфере
    unsigned int        count_blocks;
    /// \brief собственно указатель на буфер (массив блоков)
    struct block_info*  blocks;
    /// \brief текущий индекс записи
    unsigned int        index_write;
    /// \brief текущий индекс чтения
    unsigned int        index_read;
};

/// \brief The mil1553_cdev_data struct информация о символьном устройстве (канал МКИО)
struct mil1553_cdev_data {
    /// \brief device девайс
    struct device*              device;
    /// \brief cdev символьное устройство
    struct cdev                 cdev;
    /// \brief dev_major major индекс
    uint32_t                    dev_major;
    /// \brief index индекс в массиве символьных устройств
    u_int                       index;
    /// \brief numCh номер канала (minor индекс)
    uint32_t                    numCh;
    /// \brief drv ссылка на драйвер
    struct MIL1553UD_DEV*       drv;
    /// \brief devname имя устройства
    char                        devname[COUNT_NAME_SYMBOLS];
    /// \brief mtxCh мьютекс для канала
    struct mutex                mtxCh;
    /// \brief interrupt_buffer кольцевой буфер для накопления прерываний
    struct ring_buffer*         interrupt_buffer;
    /// \brief входной подадрес
    struct subaddress_data*     rcv_sa_buffer;
    /// \brief информация о сигналах для прерывания
    struct signal_info          sig_info;
};

/// \brief The mil1553_cdev_private struct приватная информация о символьном устройстве, которая прикомпоновывается к символьному устройству
struct mil1553_cdev_private {
    /// \brief index индекс в массиве символьных устройств
    u_int                       index;
};

/// \brief create_char_devs создание символьных устройств для каналов
/// \param driver_priv информационная структура об устройстве
/// \return код ошибки
int create_char_devs(struct MIL1553UD_DEV* driver_priv);

/// \brief create_char_devs удаление символьных устройств для каналов
/// \param driver_priv информационная структура об устройстве
/// \return код ошибки
int destroy_char_devs(struct MIL1553UD_DEV* driver_priv);

/// \brief getCdevPtr получить ссылку на информацию о символном устройстве (канале)
/// \param priv внутренняя структура символьного устройства
/// \return информационная структура о символьном устройстве
struct mil1553_cdev_data* getCdevPtr(struct mil1553_cdev_private *priv);

/// \brief getCdevPtrByIndex получить ссылку на информацию о символном устройстве (канале)
/// \param index индекс символьного устройства канала
/// \return информационная структура о символьном устройстве
struct mil1553_cdev_data* getCdevPtrByIndex(int index);

#endif // MIL1553UD_CDEV_H
