/// \file mil1553ud_pci.с
/// \brief инфраструктура для работы с pci устройством (pci-драйвер)
#include "mil1553ud_pci.h"
#include "mil1553ud_log.h"
#include "mil1553ud_regs.h"
#include "mil1553ud_cdev.h"
#include "mil1553ud_ioctl.h"

/// \brief инициализация pci-драйвера
static int __init mil1553ud_pci_init(void);
/// \brief деинициализация pci-драйвера
static void __exit mil1553ud_pci_exit(void);
/// \brief открытие (инициализация) очередного pci-устройства
static int mil1553ud_pci_probe(struct pci_dev *pdevice, const struct pci_device_id *pid);
/// \brief закрытие (деинициализация) очередного pci-устройства
static void mil1553ud_pci_remove(struct pci_dev *pdevice);
/// \brief обработчик прерывания pci-устройства
static irqreturn_t mil1553ud_pci_interrupt_handler(int irq, void* arg);
/// \brief инициализация кол-ва каналов pci-устройства
static void init_count_of_channels(struct MIL1553UD_DEV* driver_priv);

/// \brief pdevs глобальный массив открытых pci-устройств
struct MIL1553UD_DEV* pdevs[MAX_COUNT_BOARDS];

MODULE_LICENSE(DRIVER_LICENSE);
MODULE_DESCRIPTION(DRIVER_DESCRIPTION);
MODULE_VERSION(DRIVER_VERSION);

/// \brief регистрация функции инициализации pci-драйвера
module_init(mil1553ud_pci_init);
/// \brief регистрация функции деинициализации pci-драйвера
module_exit(mil1553ud_pci_exit);

/// \brief таблица pci-устройств
static struct pci_device_id mil1553ud_pci_id_table[] = {
    { PCI_DEVICE(VENDOR_ID_NOVOMAR,     DEVICE_ID_MPCIE1553UD1)  },
    { PCI_DEVICE(VENDOR_ID_NOVOMAR,     DEVICE_ID_MPCIE1553UD2)  },
    { PCI_DEVICE(VENDOR_ID_NOVOMAR,     DEVICE_ID_PCIE1553UDX)   },
    { PCI_DEVICE(VENDOR_ID_NOVOMAR,     DEVICE_ID_MPCIE1553UD1T) },
    { PCI_DEVICE(VENDOR_ID_NOVOMAR,     DEVICE_ID_MPCIE1553UD2T) },
    { PCI_DEVICE(VENDOR_ID_NOVOMAR,     DEVICE_ID_PCIE1553UDXT)  },
    { 0 }
};
MODULE_DEVICE_TABLE(pci, mil1553ud_pci_id_table);

/// \brief таблица операций драйвера pci-устройств
static struct pci_driver mil1553ud_pci_driver = {
    .name           = DRIVER_NAME,
    .id_table       = mil1553ud_pci_id_table,
    .probe          = mil1553ud_pci_probe,
    .remove         = mil1553ud_pci_remove,
};

static void init_count_of_channels(struct MIL1553UD_DEV* driver_priv)
{
    switch (driver_priv->deviceid) {
    case DEVICE_ID_MPCIE1553UD1:
    case DEVICE_ID_MPCIE1553UD1T: {
            uint32_t reg = ioread32(driver_priv->pbar + FPGA_BOOT_STATUS) >> 24;
            switch (reg) {
                case VAL_UD1: driver_priv->current_count_of_channels = COUNT_CHANNELS_1; break;
                default: LOG_ERROR("init_count_of_channels - can't identify FPGA_BOOT_STATUS - DEVICE_ID_MPCIE1553UD1"); break;
            }
        }
        break;
    case DEVICE_ID_MPCIE1553UD2:
    case DEVICE_ID_MPCIE1553UD2T: {
            uint32_t reg = ioread32(driver_priv->pbar + FPGA_BOOT_STATUS) >> 24;
            switch (reg) {
                case VAL_UD2: driver_priv->current_count_of_channels = COUNT_CHANNELS_2; break;
                default: LOG_ERROR("init_count_of_channels - can't identify FPGA_BOOT_STATUS - DEVICE_ID_MPCIE1553UD2"); break;
            }
        }
        break;
    case DEVICE_ID_PCIE1553UDX:
    case DEVICE_ID_PCIE1553UDXT: {// доопределение кол-ва каналов
            uint32_t reg = ioread32(driver_priv->pbar + FPGA_BOOT_STATUS) >> 24;
            switch (reg) {
            case VAL_UD1: driver_priv->current_count_of_channels = COUNT_CHANNELS_1; break;
            case VAL_UD2: driver_priv->current_count_of_channels = COUNT_CHANNELS_2; break;
            case VAL_UD4: driver_priv->current_count_of_channels = COUNT_CHANNELS_4; break;
            default: LOG_ERROR("init_count_of_channels - can't identify FPGA_BOOT_STATUS - DEVICE_ID_PCIE1553UDX"); break;
            }
        }
        break;
        default: LOG_ERROR("init_count_of_channels - can't identify deviceid"); break;
    }
    LOG_RELEASE_INFO("current_count_of_channels: %d", driver_priv->current_count_of_channels);
}

static int __init mil1553ud_pci_init(void)
{
    int error, i;
    for (i = 0; i < MAX_COUNT_BOARDS; i++)
        pdevs[i] = NULL;
    LOG_RELEASE_INFO("Init %s - version %s\n", DRIVER_DESCRIPTION, DRIVER_VERSION);
    LOG_INFO("%s\n", DRIVER_COPYRIGHT);
    error = pci_register_driver(&mil1553ud_pci_driver);
    if (error != 0) {
      LOG_ERROR("pci_register_driver error %d", error);
    }
    return error;
}

static void __exit mil1553ud_pci_exit(void)
{
    LOG_RELEASE_INFO("Exit %s - version %s\n", DRIVER_DESCRIPTION, DRIVER_VERSION);
    pci_unregister_driver(&mil1553ud_pci_driver);
}

static int mil1553ud_pci_probe(struct pci_dev* pdevice, const struct pci_device_id* pid)
{
    int bar, err, i;
    uint64_t mmio_start, mmio_len;
    uint16_t status_reg, command_reg;
    struct MIL1553UD_DEV* driver_priv;
    u32 addr_dma_lo, addr_dma_hi;
    // инициализация вспомогательной структуры данных PCI-драйвера
    driver_priv = kzalloc(sizeof(struct MIL1553UD_DEV), GFP_KERNEL);
    pci_set_drvdata(pdevice, driver_priv);
    driver_priv->pdevice = pdevice;
    // инициализация регистров управления Mil1553ud
    bar = pci_select_bars(pdevice, IORESOURCE_MEM); // получаем адрес реcурса ввода/вывода
    LOG_RELEASE_INFO("Device availale MEM BAR are 0x%x\n", bar);
    err = pci_enable_device(pdevice); // "включаем" память устройства
    if (err != 0) {
      LOG_ERROR("pci_enable_device_mem error %d", err);
      return err;
    }
    pci_set_master(pdevice);
    err = pci_request_region(pdevice, bar, DRIVER_NAME); // запрашиваем необходимый регион пямяти, с определённым ранее типом
    if (err != 0) {
      LOG_ERROR("pci_request_region error %d", err);
      return err;
    }
    driver_priv->bar = bar;
    mmio_start = pci_resource_start(pdevice, 0); // получаем адрес начала блока памяти устройства
    mmio_len = pci_resource_len(pdevice, 0); // получаем общую длину блока памяти устройства
    LOG_RELEASE_INFO("Mil1553ud device resource 0: start at 0x%lx with lenght %lu\n", mmio_start, mmio_len);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 2, 0)
    driver_priv->pbar = ioremap_cache(mmio_start, mmio_len); // мапим выделенную память к аппаратуре
#else
    driver_priv->pbar = ioremap_nocache(mmio_start, mmio_len); // мапим выделенную память к аппаратуре
#endif

    driver_priv->barlen = mmio_len; // сохраняем длину выделенной памяти
    driver_priv->pdevice = pdevice; // сохраняем указатель на pci_dev
    LOG_RELEASE_INFO("Mil1553ud device mapped resource 0x%lx to 0x%p\n", mmio_start, driver_priv->pbar);
    // читаем конфигурационные регистры PCI
    pci_read_config_word(pdevice, PCI_VENDOR_ID, &driver_priv->vendorid);
    pci_read_config_word(pdevice, PCI_DEVICE_ID, &driver_priv->deviceid);
    pci_read_config_word(pdevice, PCI_REVISION_ID, &driver_priv->revisionid);
    LOG_RELEASE_INFO("Device vid: 0x%X  pid: 0x%X revid: 0x%X\n", driver_priv->vendorid, driver_priv->deviceid, driver_priv->revisionid);
    pci_read_config_word(pdevice, PCI_STATUS, &status_reg);
    pci_read_config_word(pdevice, PCI_COMMAND, &command_reg);
    LOG_RELEASE_INFO("Device status reg: 0x%X device command reg: 0x%X\n", status_reg, command_reg);
    // сохраняем struct MIL1553UD_DEV* driver_priv в массив pci устройств
    for (i = 0; i < MAX_COUNT_BOARDS; i++) {
        if (pdevs[i] == NULL) {
            pdevs[i] = driver_priv;
            driver_priv->unit = i;
            // инициализация мьютекса
            mutex_init(&driver_priv->mtx);
            mutex_init(&driver_priv->mtxIrq);
            LOG_RELEASE_INFO("Device number %d has added to list", i);
            break;
        }
    }
    //инициализация имени pci устройства
    snprintf(driver_priv->devname, sizeof(driver_priv->devname), "%s%u" , PCI_DEVICE_NAME_TEMPLATE, driver_priv->unit);
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 0, 0)
    //set DMA mask
    if (pci_set_dma_mask(driver_priv->pdevice, DMA_BIT_MASK(64)) == 0) {
        pci_set_consistent_dma_mask(driver_priv->pdevice, DMA_BIT_MASK(64));
        LOG_RELEASE_INFO("Memory");
    }
#endif
    // настройка DMA
    driver_priv->dma_info.dma_size = DMA_SIZE;
    driver_priv->dma_info.dma_vaddr = (uint8_t*)dma_alloc_coherent(&pdevice->dev, driver_priv->dma_info.dma_size, &driver_priv->dma_info.dma_paddr, GFP_KERNEL);
    if (driver_priv->dma_info.dma_vaddr == NULL) {
      LOG_ERROR("could not allocate %lu bytes for data buffer\n", (unsigned long)driver_priv->dma_info.dma_size);
      return -1;
    }
    memset(driver_priv->dma_info.dma_vaddr, 0, driver_priv->dma_info.dma_size);
    for (i = 0; i < MAX_CHANNELS; i ++) {
      driver_priv->dma_info.ch_dma_write[i] = (uint32_t*)(driver_priv->dma_info.dma_vaddr + i * DMA_SIZE_CH);
      driver_priv->dma_info.ch_dma_read[i] = (uint32_t*)(driver_priv->dma_info.ch_dma_write[i]);
      driver_priv->dma_info.ch_dma_write_irq[i] = (uint32_t*)(driver_priv->dma_info.dma_vaddr + i * DMA_SIZE_CH);
      driver_priv->dma_info.ch_dma_read_irq[i] = (uint32_t*)(driver_priv->dma_info.ch_dma_write[i]);
    }
    LOG_RELEASE_INFO("DMA_DATA_BASE phys@%llXH virt@%p\n", (unsigned long long)driver_priv->dma_info.dma_paddr, driver_priv->dma_info.dma_vaddr);

    addr_dma_lo = ((uint64_t)driver_priv->dma_info.dma_paddr & 0xffffffff);	// low part address DMA
    addr_dma_hi = (((uint64_t)driver_priv->dma_info.dma_paddr >> 32));	// hi part address DMA
    iowrite32(addr_dma_lo, (void*) (driver_priv->pbar + DMA_DATA_BASE));
    iowrite32(addr_dma_hi, (void*) (driver_priv->pbar + DMA_DATA_BASE + 0x4));

    // определение количества каналов на плате
    init_count_of_channels(driver_priv);
    // создание символьных устройств: на каждый канал - символное устройство
    create_char_devs(driver_priv);
    // инициализация прерывания
    disable_irq(pdevice->irq);  // надо сначала принудительно выключить - нашёл на stackoverflow
    err = pci_enable_msi(pdevice); // разрешаем прерывания
    if (err != 0) {
      LOG_ERROR("pci_enable_msi error %d", err);
      return err;
    }
    enable_irq(pdevice->irq); // включаем прерывания
    err = request_irq(pdevice->irq, mil1553ud_pci_interrupt_handler, IRQF_SHARED, driver_priv->devname, driver_priv); // привязываем обработчик прерывания
    if (err != 0) {
      LOG_ERROR("request_irq error %d", pdevice->irq);
      return err;
    }
    LOG_RELEASE_INFO("IRQ %d", pdevice->irq);
    LOG_RELEASE_INFO("probe finished!");
    return 0;
}

static void mil1553ud_pci_remove(struct pci_dev *pdevice)
{
    int i;
    struct MIL1553UD_DEV* driver_priv;
    driver_priv = pci_get_drvdata(pdevice);
    LOG_RELEASE_INFO("Deinit device %d", driver_priv->unit);
    // деинициализация ДМА
    dma_free_coherent(&pdevice->dev, driver_priv->dma_info.dma_size, driver_priv->dma_info.dma_vaddr, driver_priv->dma_info.dma_paddr);
    for (i = 0; i < MAX_CHANNELS; i ++)
      driver_priv->dma_info.ch_dma_vaddr[i] = NULL;
    driver_priv->dma_info.dma_vaddr = NULL;
    driver_priv->dma_info.dma_paddr = 0;
    driver_priv->dma_info.dma_size = 0;
    // деинициализация прерывания
    free_irq(pdevice->irq, driver_priv);
    pci_disable_msi(pdevice);
    // уничтожение мьютексов
    mutex_destroy(&driver_priv->mtx);
    // уничтожение символьных устройств
    destroy_char_devs(driver_priv);
    // деинициализация вспомогательной структуры данных PCI-драйвера
    if (driver_priv) {
        if (driver_priv->pbar) {
            iounmap(driver_priv->pbar);
        }
        pdevs[driver_priv->unit] = NULL;
        kfree(driver_priv);
    }
    pci_release_region(pdevice, pci_select_bars(pdevice, IORESOURCE_MEM));
    pci_disable_device(pdevice);
    LOG_RELEASE_INFO("Unloaded %s\n", DRIVER_DESCRIPTION);
}

/// \brief макрос для генерации битовой маски в зависимости от канала
#define SWITCH_CHANNEL_BIT_MASK(numCh, ch1, ch2, ch3, ch4)     \
    (1 << ((numCh == CH_NUM_1) ? ch1 : ((numCh == CH_NUM_2) ? ch2 : ((numCh == CH_NUM_3) ? ch3 : ch4))))

/// \brief обработчик накопления прерываний канала
void inline irq_channel_handler(struct MIL1553UD_DEV* driver_priv, struct mil1553_cdev_data *cdev, struct block_info* interrupt_info);
/// \brief обработчик канала по чтению данных из дма
void inline irq_channel_dma_read_handler(struct MIL1553UD_DEV* driver_priv, struct mil1553_cdev_data *cdev, struct block_info* interrupt_info);
/// \brief обработчик канала по отправке сигнала
void inline irq_send_signal_handler(struct MIL1553UD_DEV* driver_priv, struct mil1553_cdev_data *cdev, struct block_info* interrupt_info);

static irqreturn_t mil1553ud_pci_interrupt_handler(int irq, void* arg)
{
    int i;                                                              // счётчик
    struct block_info interruptInfo[MAX_CHANNELS];                      // инфо о прерывании поканально
    unsigned int interrupt;                                             // регистр прерываний
    unsigned int interruptMask;                                         // регистр маскирования прерываний
    unsigned int regFreeTmr[MAX_CHANNELS] = { FREE_TIMER_CH1, FREE_TIMER_CH2, FREE_TIMER_CH3, FREE_TIMER_CH4 };  // адреса таймеров
    struct MIL1553UD_DEV* driver_priv = (struct MIL1553UD_DEV*)arg;     // информационная структура драйвера pci-устройства

    //mutex_lock(&driver_priv->mtxIrq);

    for (i = 0; i < driver_priv->current_count_of_channels; i++) { // чтение времени поканально
        interruptInfo[i].free_timer_value = ioread32(driver_priv->pbar + regFreeTmr[i]);
    }
    // чтение регистров прерываний
    interrupt       = ioread32(driver_priv->pbar + INTERRUPT);
    interruptMask   = ioread32(driver_priv->pbar + INTERRUPT_MASK);
    //LOG_INFO("irq = %d, interrupt = 0x%x", irq, interrupt);
    interrupt &= interruptMask; // маскируем
    if (interrupt == 0) {
        //mutex_unlock(&driver_priv->mtxIrq);
        return IRQ_NONE;
    }
    // читаем нужные регистры поканально
    for (i = 0; i < driver_priv->current_count_of_channels; i++) {
        interruptInfo[i].interrupt_ch = 0;
        if (interrupt & SWITCH_CHANNEL_BIT_MASK(i,  0,  2,  4,  6))    interruptInfo[i].interrupt_ch |= (1 << BLINF_INT_HDAT);
        if (interrupt & SWITCH_CHANNEL_BIT_MASK(i,  1,  3,  5,  7))    interruptInfo[i].interrupt_ch |= (1 << BLINF_INT_QDAT);
        if (interrupt & SWITCH_CHANNEL_BIT_MASK(i,  8, 12, 16, 20))    interruptInfo[i].interrupt_ch |= (1 << BLINF_RT_INT_SADDR);
        if (interrupt & SWITCH_CHANNEL_BIT_MASK(i,  9, 13, 17, 21))    interruptInfo[i].interrupt_ch |= (1 << BLINF_RT_INT_MC_ERR);
        if (interrupt & SWITCH_CHANNEL_BIT_MASK(i, 10, 14, 18, 22))    interruptInfo[i].interrupt_ch |= (1 << BLINF_INT_BC);
        if (interrupt & (1 << 24))                                     interruptInfo[i].interrupt_ch |= (1 << BLINF_INT_FLASH);
        if (interrupt & SWITCH_CHANNEL_BIT_MASK(i, 11, 15, 13, 23))    interruptInfo[i].interrupt_ch |= (1 << BLINT_INT_DATA_CNT_EN);
        if (interrupt & SWITCH_CHANNEL_BIT_MASK(i, 24, 25, 26, 27))    interruptInfo[i].interrupt_ch |= (1 << BLINT_TIMEOUT_ITVEN);
        if (interrupt & SWITCH_CHANNEL_BIT_MASK(i, 28, 29, 30, 31))    interruptInfo[i].interrupt_ch |= (1 << BLINT_TIMEOUT_ABSEN);
        interruptInfo[i].hw_stat_reg1 = ioread32(driver_priv->pbar + CHOOSE_CHANNEL_REG(i, HW_STAT_REG1_CH1, HW_STAT_REG1_CH2, HW_STAT_REG1_CH3, HW_STAT_REG1_CH4));
        interruptInfo[i].hw_stat_reg2 = ioread32(driver_priv->pbar + CHOOSE_CHANNEL_REG(i, HW_STAT_REG2_CH1, HW_STAT_REG2_CH2, HW_STAT_REG2_CH3, HW_STAT_REG2_CH4));
    }
    // обрабатываем поканально
    for (i = 0; i < driver_priv->current_count_of_channels; i++) {
        struct mil1553_cdev_data*  cdev = getCdevPtrByIndex(driver_priv->index_cdev[i]);  // инфо о cdev
        if (cdev == NULL) {
            LOG_ERROR("irq = %d, numCh = %d, indexCDev = %d", irq, i, driver_priv->index_cdev[i]);
            break;
        }
        //mutex_lock(&cdev->mtxCh);
        irq_channel_handler(driver_priv, cdev, &interruptInfo[i]);
        //irq_channel_dma_read_handler(driver_priv, cdev, &interruptInfo[i]);
        //mutex_unlock(&cdev->mtxCh);
        //irq_send_signal_handler(driver_priv, cdev, &interruptInfo[i]);
    }
    //mutex_unlock(&driver_priv->mtxIrq);
    return IRQ_HANDLED;
}

void inline irq_channel_handler(struct MIL1553UD_DEV* driver_priv, struct mil1553_cdev_data* cdev, struct block_info* interrupt_info)
{
    // если прерывания нет, то выходим
    if (interrupt_info->interrupt_ch == 0)
        return;
    // добавляем блок в кольцевой буфер
    cdev->interrupt_buffer->blocks[cdev->interrupt_buffer->index_write] = *interrupt_info;
    // смещаем индекс
    //mutex_lock(&cdev->mtxCh);
    cdev->interrupt_buffer->index_write++;
    if (cdev->interrupt_buffer->index_write == cdev->interrupt_buffer->count_blocks)
        cdev->interrupt_buffer->index_write = 0;
    //mutex_unlock(&cdev->mtxCh);
}

#define CALC_DMA_WRITE_CH_IRQ(drv, num_ch, dma_index)           \
    drv->dma_info.ch_dma_write_irq[num_ch] = (uint32_t*)(drv->dma_info.dma_vaddr + num_ch * DMA_SIZE_CH + dma_index * SIZE_RAW_DMA_BLOCK)

/// \brief значение адреса указателя записи для канала
#define ADR_DMA_WRITE_IRQ(drv, num_ch)            (uint64_t)(drv->dma_info.ch_dma_write_irq[num_ch])
/// \brief значение адреса указателя чтения для канала
#define ADR_DMA_READ_IRQ(drv, num_ch)             (uint64_t)(drv->dma_info.ch_dma_read_irq[num_ch])
/// \brief значение базового адреса указателя буфера ДМА
#define ADR_DMA_VADR(drv)                     (uint64_t)(drv->dma_info.dma_vaddr)

#define CALC_NBLOCK_CH_IRQ(drv, num_ch)      \
    ( \
    (ADR_DMA_WRITE_IRQ(drv, num_ch) >= ADR_DMA_READ_IRQ(drv, num_ch)) ? \
    ( (ADR_DMA_WRITE_IRQ(drv, num_ch) - ADR_DMA_READ_IRQ(drv, num_ch)) / SIZE_RAW_DMA_BLOCK ) : \
    ( (ADR_DMA_WRITE_IRQ(drv, num_ch) - (ADR_DMA_VADR(drv) + num_ch * DMA_SIZE_CH)) / SIZE_RAW_DMA_BLOCK + ((ADR_DMA_VADR(drv) + num_ch * DMA_SIZE_CH + MEGABYTE) - ADR_DMA_WRITE_IRQ(drv, num_ch)) / SIZE_RAW_DMA_BLOCK ) \
    )

/// \brief значение адреса конца буфера дма для канала
#define ADR_DMA_CH_END(drv, num_ch)  \
    ADR_DMA_VADR(drv) + num_ch * DMA_SIZE_CH + DMA_SIZE_CH

void inline rt_dma_handler(void* dma_block, struct mil1553_cdev_data *cdev);
void inline bc_dma_handler(void* dma_block, struct mil1553_cdev_data *cdev);

void inline irq_channel_dma_read_handler(struct MIL1553UD_DEV* driver_priv, struct mil1553_cdev_data *cdev, struct block_info* interrupt_info)
{
    unsigned int nblock;        // кол-во блоков, доступных для чтения
    unsigned int nblockend;     // кол-во блоков, до границы буфера
    unsigned int dma_index;     // индекс дма из регистра DMA_INDEX
    unsigned char dma_block[SIZE_RAW_DMA_BLOCK];    // блок дма
    // если прерывания нет, то выходим
    if ( ((interrupt_info->interrupt_ch & BLINF_RT_INT_SADDR) == 0) || ((interrupt_info->interrupt_ch & BLINF_INT_BC) == 0) )
        return;
    dma_index = (ioread32(driver_priv->pbar + CHOOSE_CHANNEL_REG(cdev->numCh, DMA_INDEX_CH1, DMA_INDEX_CH2, DMA_INDEX_CH3, DMA_INDEX_CH4)) >> 7) & 0x1FFF;
    CALC_DMA_WRITE_CH_IRQ(driver_priv, cdev->numCh, dma_index);
    nblock = CALC_NBLOCK_CH_IRQ(driver_priv, cdev->numCh);
    nblockend = (ADR_DMA_CH_END(driver_priv, cdev->numCh) - ADR_DMA_READ_IRQ(driver_priv, cdev->numCh)) / SIZE_RAW_DMA_BLOCK; // вычисляем кол-во блоков, до границы буфер
    // обрабатываем блоки дма
    while ( ADR_DMA_WRITE_IRQ(driver_priv, cdev->numCh) != ADR_DMA_READ_IRQ(driver_priv, cdev->numCh) ) {
        // копируем блок ДМА во временную переменную
        memcpy( (void*)dma_block, (void*)driver_priv->dma_info.ch_dma_read_irq[cdev->numCh], SIZE_RAW_DMA_BLOCK );
        // смещаем на блок
        driver_priv->dma_info.ch_dma_read_irq[cdev->numCh] += SIZE_RAW_DMA_BLOCK / 4;
        if (ADR_DMA_READ_IRQ(driver_priv, cdev->numCh) > ADR_DMA_CH_END(driver_priv, cdev->numCh))
            driver_priv->dma_info.ch_dma_read_irq[cdev->numCh] = (uint32_t*)(driver_priv->dma_info.dma_vaddr + cdev->numCh * DMA_SIZE_CH); // при необходимости корректируем
        if (interrupt_info->interrupt_ch & BLINF_RT_INT_SADDR) { // обработка в режиме ОУ
            rt_dma_handler((void*)dma_block, cdev);
        }
        if (interrupt_info->interrupt_ch & BLINF_INT_BC) {// обработка в режиме КШ
            bc_dma_handler((void*)dma_block, cdev);
        }
    }
}

#define DWORD_SIZE                  4
#define RT_DMA_CNT_SERVICE_WORDS    5
#define RT_DMA_NUM_SRV_WRD(num)     (num - 1)

void inline rt_dma_handler(void* dma_block, struct mil1553_cdev_data *cdev)
{
    int i;  // счётчик
    unsigned int serviceDwords[RT_DMA_CNT_SERVICE_WORDS];       // сервисные слова блока ДМА ОУ
    unsigned int rt_format;                                     // формат транзакции
    memcpy( (void*)serviceDwords, (void*)dma_block, DWORD_SIZE * RT_DMA_CNT_SERVICE_WORDS);
    // разбор блока и запись в буфер входных поддадресов
    if (serviceDwords[RT_DMA_NUM_SRV_WRD(1)] & (1 << 26))   // ошибка в принятом сообщении
        return;
    rt_format = (serviceDwords[RT_DMA_NUM_SRV_WRD(1)] >> 26) & 0xF; // получаем формат транзакции
    if ( (rt_format == 1) || (rt_format == 7) ) {
        char isLeft;            // признак младшей части
        unsigned int rt_srv_ct; // кол-во служебных слов
        unsigned int rt_dat_ct; // кол-во сд
        unsigned int* dataWords;    // указатель на слова данных
        unsigned int subaddr;       // поддадрес приёма
        rt_srv_ct = (serviceDwords[RT_DMA_NUM_SRV_WRD(1)] >> 10) & 0xF;
        rt_dat_ct = (serviceDwords[RT_DMA_NUM_SRV_WRD(1)] >> 4) & 0xF;
        dataWords = ((unsigned int*)dma_block) + RT_DMA_CNT_SERVICE_WORDS + rt_srv_ct / 2;
        if ( rt_srv_ct % 2 )
            dataWords++;
        isLeft = 1;
        subaddr = serviceDwords[RT_DMA_NUM_SRV_WRD(1)] & 0x1F;
        for (i = 0; i < rt_dat_ct; i++) {
            if (isLeft) {
                cdev->rcv_sa_buffer->data[subaddr-1][i] = *dataWords & 0xFFFF;
                isLeft = 0;
            } else {
                cdev->rcv_sa_buffer->data[subaddr-1][i] = (*dataWords >> 16) & 0xFFFF;
                isLeft = 1;
                dataWords++;
            }
        }
    }
}

#define BC_DMA_NUM_SRV_WRD(num)     (num - 1)

void inline bc_dma_handler(void* dma_block, struct mil1553_cdev_data *cdev)
{
    int i;  // счётчик
    unsigned int serviceDwords[RT_DMA_CNT_SERVICE_WORDS];       // сервисные слова блока ДМА КШ
    unsigned int rt_format;                                     // формат транзакции
    memcpy( (void*)serviceDwords, (void*)dma_block, DWORD_SIZE * RT_DMA_CNT_SERVICE_WORDS);
    // разбор блока и запись в буфер входных поддадресов
    if ( (serviceDwords[BC_DMA_NUM_SRV_WRD(1)] & (1 << 28)) == 0 )   // ошибка в принятом сообщении
        return;
    rt_format = (serviceDwords[BC_DMA_NUM_SRV_WRD(1)] >> 1) & 0xF;
    if (rt_format == 2) {
        char isLeft;            // признак младшей части
        unsigned int rt_srv_ct; // кол-во служебных слов
        unsigned int rt_dat_ct; // кол-во сд
        unsigned int* dataWords;    // указатель на слова данных
        unsigned int subaddr;       // поддадрес приёма
        rt_srv_ct = ((serviceDwords[BC_DMA_NUM_SRV_WRD(1)] >> 11) & 0xF) - 5*2;
        rt_dat_ct = ((serviceDwords[BC_DMA_NUM_SRV_WRD(1)] >> 5) & 0xF) & 0x3F;
        if (rt_dat_ct > 31) rt_dat_ct = 32;
        dataWords = ((unsigned int*)dma_block) + RT_DMA_CNT_SERVICE_WORDS;
        subaddr = ((*dataWords & 0xFFFF) >> 5) & 0x1F;
        dataWords += 3; // переходим к словам данных
        isLeft = 1;
        for (i = 0; i < rt_dat_ct; i++) {
            if (isLeft) {
                cdev->rcv_sa_buffer->data[subaddr-1][i] = *dataWords & 0xFFFF;
                isLeft = 0;
            } else {
                cdev->rcv_sa_buffer->data[subaddr-1][i] = (*dataWords >> 16) & 0xFFFF;
                isLeft = 1;
                dataWords++;
            }
        }
    }
}



